/*************************************************************************
    main.c -- last change: 11-11-1996

    Multitrack is a program to play, record and manipulate tracks. The
    tracks are stored on disk.

    Copyright (C) 1996  Boris Nagels

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    Sent your comments and suggestions to Boris Nagels, E-mail:
    boris@molphys.leidenuniv.nl

 *************************************************************************/


#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>
#include <stdlib.h>

/* For shared memory */
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>

#include <vga.h>
#include "multitrack.h"

/* Process id's of the childs */
pid_t pid_play = 0, pid_record = 0;

/* Global definitions for the shared memory blocks */
Track *track;
Track_info *info;
short int *bigbuf;
Buffer_info *bigbuf_info;
int track_memid, info_memid, buf_memid, bufinfo_memid;

/* Global definitions for screen functions */
Button button[nr_buttons];
Slider slider[nr_sliders];
Screen screen[nr_tracks];
int mouse_x, mouse_y, mouse_moved, sub_window = FALSE;
int select_begin = screen_min, select_end = screen_min;
int screen_length = screen_max - screen_min + 1;
int play_position, old_play_position;
int edit_track = -1;
char *save_screen;
void *font;

/* Global definitions for tracks */
long track_begin, track_end, track_length;

/* Global definition for the soundcards */
Soundcard soundcard[2];


static void term_handler(int sig)
{
	vga_setmode(TEXT);	/* switch to textmode */

	fprintf(stderr, "Killing childs ...\n");
	signal(SIGCHLD, SIG_DFL);	/* deflag child signal handler */
	if (pid_play) {
		kill(pid_play, SIGKILL);	/* kill the childs */
		waitpid(pid_play, NULL, WNOHANG);
	}
	if (pid_record) {
		kill(pid_record, SIGKILL);
		waitpid(pid_record, NULL, WNOHANG);
	}
	fprintf(stderr, "Detaching shared memory ...\n");
	if (shmdt((char *) track) < 0)
		perror("Cannot detach shared memory for track");
	if (shmdt((char *) info) < 0)
		perror("Cannot detach shared memory for info");
	if (shmdt((char *) bigbuf) < 0)
		perror("Cannot detach shared memory for bigbuf");
	if (shmdt((char *) bigbuf_info) < 0)
		perror("Cannot detach shared memory for bigbuf_info");

	fprintf(stderr, "Goodbye.\n");
	exit(1);
}


static void chld_handler(int sig)
{
	fprintf(stderr, "My child died ...\n");
	waitpid(pid_play, NULL, WNOHANG);
	waitpid(pid_record, NULL, WNOHANG);
	term_handler(0);
}


void error(char *s)
{
	fprintf(stderr, "Error: %s\n", s);
	term_handler(0);
	exit(1);
}


static void main_parent(void)
{
	gui();
}


static void allocate_shared_mem(void)
{
	track_memid = shmget(IPC_PRIVATE, sizeof(Track) * nr_tracks,
			     IPC_EXCL | IPC_CREAT | 0666);
	if (track_memid == -1)
		error("Can't get shared memory for track");
	track = (Track *) shmat(track_memid, (char *) 0, 0);
	if (track == (Track *) - 1)
		error("Can't attach shared memory for track");


	info_memid = shmget(IPC_PRIVATE, sizeof(Track_info),
			    IPC_EXCL | IPC_CREAT | 0666);
	if (info_memid == -1)
		error("Can't get shared memory for info");
	info = (Track_info *) shmat(info_memid, (char *) 0, 0);
	if (info == (Track_info *) - 1)
		error("Can't attach shared memory for info");


	buf_memid = shmget(IPC_PRIVATE, sizeof(short int) * 22050 * 24,
			   IPC_EXCL | IPC_CREAT | 0666);
	if (buf_memid == -1)
		error("Can't get shared memory for bigbuf");
	bigbuf = (short int *) shmat(buf_memid, (char *) 0, 0);
	if (bigbuf == (short int *) -1)
		error("Can't attach shared memory for bigbuf");


	bufinfo_memid = shmget(IPC_PRIVATE, sizeof(Buffer_info),
			       IPC_EXCL | IPC_CREAT | 0666);
	if (bufinfo_memid == -1)
		error("Can't get shared memory for bigbuf_info");
	bigbuf_info = (Buffer_info *) shmat(bufinfo_memid, (char *) 0, 0);
	if (bigbuf_info == (Buffer_info *) - 1)
		error("Can't attach shared memory for bigbuf_info");

	fprintf(stderr, "Marking shared memory ...\n");
	if (shmctl(track_memid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		perror("Cannot remove shared memory for track");
	if (shmctl(info_memid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		perror("Cannot remove shared memory for info");
	if (shmctl(buf_memid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		perror("Cannot remove shared memory for bigbuf");
	if (shmctl(bufinfo_memid, IPC_RMID, (struct shmid_ds *) 0) < 0)
		perror("Cannot remove shared memory for bigbuf_info");
}


void main(int argc, char *argv[])
{
	if (argc > 1)
		help();

	signal(SIGTERM, term_handler);
	signal(SIGINT, term_handler);
	signal(SIGSEGV, term_handler);
	signal(SIGCHLD, chld_handler);

	allocate_shared_mem();

	load_RC();
	init_soundcards();
	init_mixers();

	pid_play = fork();
	if (pid_play == -1)
		error("Cannot create play-child.");
	if (pid_play == 0)	/* this is the play-child */
		main_playchild();
	else {			/* this is the parent */
		pid_record = fork();
		if (pid_record == -1)
			error("Cannot create record-child.");
		if (pid_record == 0)	/* this is the record-child */
			main_recordchild();
		else
			main_parent();
	}
	term_handler(0);	/* Clean up shared memory and kill childs */
}
