/*************************************************************************
    play.c -- last change: 11-11-1996

    Multitrack is a program to play, record and manipulate tracks. The
    tracks are stored on disk.

    Copyright (C) 1996  Boris Nagels

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
    Sent your comments and suggestions to Boris Nagels, E-mail:
    boris@molphys.leidenuniv.nl

 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include "multitrack.h"

#define read_length 4096

/* Local variables for soundcard */
static int audio_fd, effect_fd;
static int mixer_fd;
static int max_slicelen;

/* Local variables for output */
static short int *data[nr_tracks];
static short int *output, *output_effect;
static int fd[nr_tracks], mute[nr_tracks], record[nr_tracks];
static long buffer_offset, extra_samples, mix_position;
static int play_buffer;

static void error_play(char *s)
{
	fprintf(stderr, "Error play: %s\n", s);
	exit(0);
}


static void analyse_buttons(void)
{
	int pfl = FALSE, count;
	long position;

	position = info->time_count;
	if (info->ff && position + 2 * read_length < info->play_end) {
		extra_samples += 2 * read_length;
		for (count = 0; count < nr_tracks; count++)
			lseek(fd[count], 2 * read_length * 2, SEEK_CUR);
	}
	if (info->rew) {
		position -= 3 * read_length;
		if (position < info->play_begin) {
			position = info->play_begin;
			extra_samples -= read_length;
		} else
			extra_samples -= 3 * read_length;
		for (count = 0; count < nr_tracks; count++)
			lseek(fd[count], position * 2, SEEK_SET);
	}
	for (count = 0; count < nr_tracks; count++) {
		mute[count] = track[count].mute;
		if (track[count].pfl)
			pfl = TRUE;
	}
	if (pfl)
		for (count = 0; count < nr_tracks; count++) {
			if (track[count].pfl)
				mute[count] = FALSE;
			else
				mute[count] = TRUE;
		}
}


static void mix_slice(int slicelen, int offset)
{
	int count, t;
	short int *inptr, *outptr;

	for (count = 0; count < nr_tracks; count++) {
		if (!(mute[count] || record[count])) {
			outptr = output;
			inptr = data[count] + offset;
			for (t = slicelen; t; t--) {
				*(outptr++) += (*(inptr) * track[count].volume_r) >> 8;
				*(outptr++) += (*(inptr++) * track[count].volume_l) >> 8;
			}
		}
	}
}


static void mix_buffer_slice(int slicelen)
{
	int t;
	short int *inptr, *outptr;

	outptr = output;
	inptr = bigbuf + buffer_offset;
	for (t = slicelen; t; t--) {
		*(outptr++) += (*(inptr) * bigbuf_info->volume) >> 8;
		*(outptr++) += (*(inptr++) * bigbuf_info->volume) >> 8;
	}
}


static void mix_effect_slice(int slicelen, int offset)
{
	int count, t;
	short int *inptr, *outptr;

	for (count = 0; count < nr_tracks; count++) {
		if (track[count].volume_effect) {
			outptr = output_effect;
			inptr = data[count] + offset;
			for (t = slicelen; t; t--)
				*(outptr++) += (*(inptr++) * track[count].volume_effect) >> 8;
		}
	}
}


static void update_time(void)
{
	count_info sndcrd_info;

	if (soundcard[0].ok)
		ioctl(audio_fd, SNDCTL_DSP_GETOPTR, &sndcrd_info);
	else
		sndcrd_info.bytes = 0;

	info->time_count = info->play_begin + sndcrd_info.bytes / 4 +
	    extra_samples;
}


static void play_slice(int number)
{
	int loop_position = 0, slicelen;

	do {
		slicelen = number - loop_position;
		if (slicelen > max_slicelen)
			slicelen = max_slicelen;

		update_time();
		bzero((char *) output, max_slicelen * 2 * nr_outputs);
		mix_slice(slicelen, loop_position);
		if (info->effect_send && !info->recording) {
			bzero((char *) output_effect, max_slicelen * 2);
			mix_effect_slice(slicelen, loop_position);
		}
		loop_position += slicelen;

		mix_position += slicelen;
		update_time();

		if (mix_position + info->play_begin >= bigbuf_info->position &&
		    buffer_offset < bigbuf_info->length && play_buffer) {
			mix_buffer_slice(slicelen);
			buffer_offset += slicelen;
			update_time();
		}
		if (soundcard[0].ok)
			write(audio_fd, (char *) output, slicelen * 2 * nr_outputs);
		if (soundcard[1].ok && info->effect_send && !info->recording)
			write(effect_fd, (char *) output_effect, slicelen * 2);
	}
	while (slicelen == max_slicelen && info->playing);
}


static void play(void)
{
	int count, number;
	long slicelen;
	char hulp[30];

	for (count = 0; count < nr_tracks; count++)
		lseek(fd[count], info->play_begin * 2, SEEK_SET);

	if (soundcard[0].ok) {
		open_audio(&audio_fd, 0, O_WRONLY, TRUE);
		sprintf(hulp, "/dev/mixer%d", soundcard[0].device);
		mixer_fd = open(hulp, O_RDWR);
		mute_mixer(mixer_fd, FALSE);	/* unmute the mixer */
	}
	if (info->effect_send && soundcard[1].ok && !info->recording)
		open_audio(&effect_fd, 1, O_WRONLY, FALSE);

	play_buffer = FALSE;
	for (count = 0; count < nr_tracks; count++)
		record[count] = track[count].rec;

	if (bigbuf_info->show_buffer != -1)
		play_buffer = TRUE;
	number = 0;
	buffer_offset = 0;
	extra_samples = 0;	/* used for ff and rew */
	mix_position = 0;
	if (info->play_begin >= bigbuf_info->position)
		buffer_offset = info->play_begin - bigbuf_info->position;
	do {
		analyse_buttons();
		slicelen = (info->play_end - info->play_begin) - mix_position - extra_samples;
		if (slicelen > read_length)
			slicelen = read_length;

		number = slicelen;
		for (count = 0; count < nr_tracks; count++)
			if (!record[count]) {
				if (!mute[count])
					number = read(fd[count], (char *) data[count], slicelen * 2) / 2;
				else
					lseek(fd[count], slicelen * 2, SEEK_CUR);
			}
		play_slice(number);
	}
	while (slicelen == read_length && info->playing);

	if (info->playing && soundcard[0].ok) {		/* wait till playing is finished */
		while (info->time_count < info->play_end - max_slicelen)
			update_time();
	}
	if (soundcard[0].ok) {
		mute_mixer(mixer_fd, TRUE);	/* mute mixer */
		ioctl(audio_fd, SNDCTL_DSP_POST);
		close(audio_fd);
		close(mixer_fd);
	}
	if (info->effect_send && soundcard[1].ok && !info->recording) {
		ioctl(effect_fd, SNDCTL_DSP_POST);
		close(effect_fd);
	}
	info->playing = FALSE;
	info->finished_playing = TRUE;
}


static void play_handler(int sig)
{
	if (sig == SIGUSR1) {
		info->playing = TRUE;
		signal(SIGUSR1, play_handler);	/* reinstall handler */
	}
	if (sig == SIGUSR2) {
		info->playing = FALSE;
		signal(SIGUSR2, play_handler);	/* reinstall handler */
	}
	return;
}


static void kill_handler(int sig)
{
	int count;

	fprintf(stderr, "Terminating play child...\n");
	for (count = 0; count < nr_tracks; count++) {
		free(data[count]);
		close(fd[count]);
	}
	free(output);
	free(output_effect);
	exit(0);
}


void main_playchild(void)
{
	char filename[80];
	int count;

	if (soundcard[0].buffersize > read_length)
		max_slicelen = read_length;
	else
		max_slicelen = soundcard[0].buffersize;

	for (count = 0; count < nr_tracks; count++) {
		sprintf(filename, "track%d.wav", count);
		fd[count] = open(filename, O_RDONLY);
		if (fd[count] == -1)
			error_play("Cannot open all tracks.");
		/* Allocate read buffer for each track (16 bit integers). */
		data[count] = (short int *) malloc(read_length * sizeof(short int));
		if (data[count] == NULL)
			error_play("Unable to allocate memory for the tracks.\n");
	}
	output = (short int *) malloc(max_slicelen * sizeof(short int) * 2);
	output_effect = (short int *) malloc(max_slicelen * sizeof(short int));
	if (output == NULL || output_effect == NULL)
		error_play("Unable to allocate memory for the output buffer.\n");

	signal(SIGUSR1, play_handler);
	signal(SIGUSR2, play_handler);
	signal(SIGKILL, kill_handler);
	while (1) {
		pause();
		if (info->playing)
			play();
	}
}
